(**************************************************************************)
(*  Burfiks: bayesian filter for the web                                  *)
(*                                                                        *)
(*  (c) Copyright 2007, INRIA                                             *)
(*                                                                        *)
(*  Authors:                                                              *)
(*  --------                                                              *)
(*  Francesco Zappa Nardelli (INRIA, Rocquencourt)                        *)
(*  Patala Arun Kumar (IIT Kanpur)                                        *)
(*  Sanchit Garg (IIT Bombay)                                             *)
(*  Vidyut Ghosal (IIT Kanpur)                                            *)
(*                                                                        *)
(*  This file is distributed under the terms of the GNU Public License    *)
(*  version 2, http://www.gnu.org/licenses/gpl-2.0.txt                    *)
(*                                                                        *)
(**************************************************************************)

(* Some notes:
   - login/passwd/profiles information are stored in the 
   "profiles/burfiks_login/" 
   directory, where the login stringis replaced with the profile login 
   (a directory for each profile).  Inside the directory, we find the 
   following files:

   passwd:
   login\n
   passwd\n
   
   profile1/
   class1
   ...
   classn

   ...

   profilen/
   class1
   ...
   classn

   the profiles are stored as directories, containing each the
   different classifier states.  

   for the time being, the profile directories are stored in the same
   directory from where the server is launched.

   - the passwords are not stored, only their md5 hash is.

   - * a new session begins with the sidebar sending the login and passwd
 * the server checks the validity of the passwd, sending back ack or failure
 * the server loads all the user profiles, sending their names back
 * the sidebar selects a profile (this may change during the time)
 * the rank function takes into account the current profile when ranking 
   a link
 * profiles can be updated (depend on the classifier) 
   
   - if the sidebar sends logout, the session is terminated, the profiles 
   are saved, and the thread responsible for the session terminates gracefully 
 *)

open Auxil

type passwd = Digest.t
type login = string

let profile_directory login public=
 if (public) then
  (let temp_name = Filename.concat (Filename.concat  !State.burfiks_dir "public") login in
    (*(Filename.concat "profiles" ("burfiks_"^login)) in*)
  temp_name)
  else
  (let temp_name = Filename.concat !State.burfiks_dir login in
    (*(Filename.concat "profiles" ("burfiks_"^login)) in*)
  temp_name)

let unlink_noerror name =
  try Unix.unlink name with
    Unix.Unix_error _ ->
      (warning("ERROR IN rmdir_noerror "^name);)
	
let rmdir_noerror name =
  try Unix.rmdir name with
    Unix.Unix_error _ ->
      (warning("ERROR IN rmdir_noerror "^name);)

let list_dir name =
  try
    let dir = Unix.opendir name in
    let rec collect l =
(*print_endline("in list_dir with "^name);*)
      let name' =
        try Some (Unix.readdir dir) with
        | Unix.Unix_error _ | End_of_file -> None
      in
      match name' with
      | Some name' -> 
	  print_endline("chking "^name');
	  if (name' <> ".") && (name' <> "..") then collect (Filename.concat name name' :: l) else collect l
      | None ->
          l
    in
    let files = collect [] in
    Unix.closedir dir;
    files
  with
  | Unix.Unix_error _ | Not_found -> []

	
(* * Delete a directory recursively. * *)
let rmdir_recursive_noerror name =

  let rec rmdir dirname =
    let rm filename =
      let stat =
        try Some (Unix.lstat filename) with
          Unix.Unix_error _ -> 
	  None
      in

      match stat with
        Some { Unix.st_kind = Unix.S_DIR } ->
          rmdir filename
      | Some _ ->
          unlink_noerror filename
      | None ->
          ()
    in

    List.iter rm (list_dir dirname);
    rmdir_noerror dirname
  in

  rmdir name

let check_login login passwd =
  let profile_directory = profile_directory login false in
  let passwd_file = Filename.concat profile_directory "passwd" in
  let hashed_passwd = Digest.string passwd in
  try 
    let in_channel = open_in passwd_file in
    let in_login = input_line in_channel in
    let in_passwd = input_line in_channel in
    let succes =
      if (not ((compare in_login login) = 0))
      then ( 
	warning ("corrupted profile: "^login); 
	warning ("expected -"^login^"- obtained -"^in_login^"-");
	false )
      else if (not ((compare in_passwd hashed_passwd) = 0))
      then (
	warning ("invalid password");
	false )
      else 
	true
    in 
    close_in in_channel;
    succes
  with Sys_error _ -> (
    warning ("user does not exists: "^login);
    false )

let create_user login passwd =
  try
    let publicprofile_directory = profile_directory login true in
    let profile_directory = profile_directory login false in
    Unix.mkdir publicprofile_directory 0o700;
    Unix.mkdir profile_directory 0o700;
    let passwd_file = Filename.concat profile_directory "passwd" in
    let hashed_passwd = Digest.string passwd in
    try
      let out_channel = open_out passwd_file in
      output_string out_channel (login^"\n");
      output_string out_channel (hashed_passwd^"\n");
      close_out out_channel;
      true
    with Sys_error _ -> (
      warning ("could not write passwd file");
      false )
  with _ -> (
    warning ("could not create user"); 
    false )

let create_profile login name = 
  try
    let profile_directory = profile_directory login false in
    let profile_path = Filename.concat profile_directory (name^".db") in

    (*Unix.mkdir profile_path 0o700;*)
    (*Unix.mkdir profile_directory 0o700;*)
let stat =
        try Some (Unix.lstat profile_path) with
          Unix.Unix_error _ -> None in

    match stat with
       Some _ -> false
      | None -> true
  with _ -> (
    warning ("could not create profile"); 
    false )

let delete_profile login name = 
  try 
    let profile_direct = profile_directory login false in
    (*let profile_path = Filename.concat profile_direct name in*)
    (*Unix.rmdir*) 
    (*rmdir_recursive_noerror profile_path;*)
    let profile_path = Filename.concat profile_direct (name^".db") in
    Unix.unlink profile_path;
    true
  with _ -> (
    warning ("could not delete profile"); 
    false )

let delete_publicprofile login name = 
  try 
    let profile_direct = profile_directory login true in
    let profile_path = Filename.concat profile_direct (name^".db") in
    Unix.unlink profile_path;
    true
  with _ -> (
    warning ("could not delete profile"); 
    false )

let move_profile login name =
  try
   let name = (name^".db")  in
   let newname = Filename.concat (profile_directory login true) name in
   let oldname= Filename.concat (profile_directory login false) name in
     Unix.rename oldname newname;
   true
   with _ -> (
    warning ("could not move profile"); 
    false)

let unmove_profile login name =
  try
   let name = (name^".db")  in
   let oldname = Filename.concat (profile_directory login true) name in
   let newname= Filename.concat (profile_directory login false) name in
     Unix.rename oldname newname;
   true
   with _ -> (
    warning ("could not move profile"); 
    false)

let delete_user login = 
  try
    let prof_directory = profile_directory login false in
    rmdir_recursive_noerror prof_directory ;
    let publicprofile_directory = profile_directory login true in
    rmdir_recursive_noerror publicprofile_directory ;
    true
  with _ -> (
    warning ("could not delete user"); 
    false )


let collect_privateprofiles login =
  let handle = Unix.opendir (profile_directory login false) in
  let profile_list = ref "" in
  try
    let skip_dir = [ ".."; "."; "passwd" ] in
    while (true) do
      let profile_name = Unix.readdir handle in
      if (not (List.mem profile_name skip_dir)) then
	profile_list := 
	  !profile_list 
	  ^ "EXISTING_PROFILE PRIVATE "^login^" "^ (String.sub profile_name 0 (String.length profile_name -3)) ^ " \n";
    done; Unix.closedir(handle) ; ""
  with _ -> !profile_list

let collect_publicprofiles login =
  let handle = Unix.opendir (profile_directory login true) in
  let profile_list = ref "" in
  try
    let skip_dir = [ ".."; "."; "passwd" ] in
    while (true) do
      let profile_name = Unix.readdir handle in
      if (not (List.mem profile_name skip_dir)) then
	profile_list := 
	  !profile_list 
	  ^ "EXISTING_PROFILE PUBLIC "^login^" "^ (String.sub profile_name 0 (String.length profile_name -3)) ^ " \n";
    done; Unix.closedir(handle) ; ""
  with _ -> !profile_list


let collect_profiles login =
  let handle = Unix.opendir (profile_directory "public" false) in
  let user_list = ref (collect_privateprofiles login) in
  try
    let skip_dir = [ ".."; "."; "passwd" ] in
    while (true) do
      let user_name = Unix.readdir handle in
      if (not (List.mem user_name skip_dir)) then
	user_list := 
	  !user_list 
	  ^ (collect_publicprofiles user_name);
    done; Unix.closedir(handle) ; ""
  with _ -> !user_list^"END_EXISTING_PROFILES\n"













